#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "lexer.h"

/*******************************
I'll make a generic lexer here, to be used in my projects.
I'll modify it to suit the purpose but most of the time, I'm just
using the same lexemes over and over. Here we go:

 identifier: letter (letter|digit)+
 quoted : quote quotchar quote
 quotchar: (backslash anychar) | (everthingexceptquote)
 decimalnumber: digits [ . digits [(E|e)[+-]digits ]
 octalnumber: 0 octdigit+
 hexnumber: 0(x|X) hexdigit+
 binnumber: 0(b|B) bindigit+
  { } ( ) / + - * ! = : $ % @  ^ & ? < >  

The output of the lexer will be as follows:

 LEtok  : token number 
 LEsub  : subclass of the token
 LEval  : value
 LElin  : line number
 LEfil  : file name

It will be possible to switch to different modes, for example a C-code
mode etc.

It will also be possible to parse a file or a memory buffer.

*****************************************************************/

int LEtok, LEsub, LElin;
unsigned char *LEval;
char *LEfil;
static unsigned int LEvalsize, LEvallen;
static unsigned char *LEinp;
static unsigned int LEinpsize, LEinplen;
static int LEmethod;
static int LEinpH; // input-head
static int LEinpOFF; // offset of the current input buffer
                     // with regards to the whole input consumed
static int LEtoks;   // start of the current token
static int LEtokl;   // line number of the start of current token

static FILE *LEfilep;

static unsigned char LEcls[256];

enum { LEclsDIGIT=1, LEclsSPACE, LEclsLETTER, LEclsOPER, LEclsQUOTE,
       LEclsCOMMENT, LEclsNEWLINE, LEclsBAD };


static void LEclsSETr(int s,int e, int v)
{
  int i;
  for(i=s;i<=e;i++) LEcls[i]= v;
}


void LEinit()
{
#ifndef LEbufSIZE   
#define LEbufSIZE   256
#endif
  LEval= malloc(LEbufSIZE);
  LEvalsize= LEbufSIZE;
  LEvallen= 0;

  if (!LEcls[0])
  {
    LEclsSETr(0,127, LEclsOPER);
    LEclsSETr(128,255, LEclsBAD);

    LEclsSETr(0,32, LEclsSPACE);
    LEcls[0x7f]= LEclsSPACE;
    LEcls[0x0a]= LEclsNEWLINE;

    LEclsSETr('a', 'z', LEclsLETTER);
    LEclsSETr('A', 'Z', LEclsLETTER);
    LEcls['_']= LEclsLETTER;

    LEclsSETr('0','9', LEclsDIGIT);
    LEcls['\"']= LEclsQUOTE;
    LEcls['\'']= LEclsQUOTE;
    LEcls['#']= LEclsCOMMENT;
  }
}

int LEfile(char *fn)
{
#ifndef LEinpSIZE   
#define LEinpSIZE   256
#endif
  if (LEinp) LEclose();

  LEfilep= fopen(fn,"rb");
  if (!LEfilep) return 1;
  
  LEinp= malloc(LEinpSIZE);
  LEinpsize= LEinpSIZE;

  LEinpH= 0;
  LEinplen= 0;
  LEmethod= 'f';
  LEfil= strdup(fn);
  return 0;
}

static int LEget(int advance)
{
  if (advance) LEinpH++;
  if (LEinpH>=LEinplen)
  {
    if (LEmethod!='f' || !LEfilep) return 1;
    LEinpOFF+= LEinpsize;
    LEinplen= fread(LEinp, 1, LEinpsize, LEfilep);
    if (LEinplen<=0) return 1;
    LEinpH= 0;
  }
  return 0;
}

void LEclose()
{
  if (LEmethod=='f')
  {
    if (!LEfilep) return ;
    fclose(LEfilep);
    LEfilep= NULL;
    free(LEinp);
  }
  LEmethod= 0;
  LEinp= NULL;
  LEinplen= 0;
  LEinpH= 1;
}

void LEmemory(unsigned char *buf, int len,char *name)
{
  if (LEinp) LEclose();

  LEinp= buf;
  LEinplen= len;
  LEmethod= 'm';
  LEinpH= 0;
  LElin= 0;
  if (name) LEfil= strdup(name);
       else LEfil= strdup("internal-source");
}

static void LEput(int X)
{
again:
  if (LEvallen<LEvalsize)
  {
    LEval[LEvallen++]= X;
    return ;
  }
  LEvalsize += LEvalsize/2 + 20;
  LEval= realloc(LEval, LEvalsize);
  goto again;
}

static int LEtoken(int tok)
{
  LEput(0);
  LEtok= tok;
  return tok;
}

static int hexval(int A)
{
  if (A>='0' && A<='9') return A-'0';
  if (A>='a' && A<='z') return A-'a'+10;
  if (A>='A' && A<='Z') return A-'A'+10;
  return -1;
}


int LEnext()
{
#define LEc  LEinp[LEinpH]
  int d1,d2, q;
  LEvallen= 0;
again:
  LEtoks= LEinpOFF + LEinpH;
  LEtokl= LElin;
  if (LEget(0)) return LEtoken(LEtEOF);
  switch(LEcls[LEc])
  {
  case LEclsSPACE:   LEinpH++; goto again;
  case LEclsNEWLINE: LEinpH++; LElin++; goto again;
  case LEclsCOMMENT: goto Tcomment;
  case LEclsQUOTE:   goto Tquote;
  case LEclsOPER:    goto Toper;
  case LEclsLETTER:  goto Tident;
  case LEclsDIGIT:   goto Tnumber;
  default:
     return LEtoken(LEtBADCHAR);
  }

Tcomment:
  if (LEget(1)) return LEtoken(LEtEOF);
  if (LEc==0x0a) goto again;
  goto Tcomment;
Tquote:
  q= LEc;
  LEsub= q;
quotcont:
  if (LEget(1)) return LEtoken(LEtUNTERMQ);
  if (LEc==q)
  {
    LEinpH++;
    LEput(0);
    return LEtoken(LEtQUOTED);
  }
  if (LEc=='\\')
  {
    if (LEget(1)) return LEtoken(LEtUNTERMQ);
    switch(LEc)
    {
    case 'q': LEput('\''); break;
    case 'd': LEput('\"'); break;
    case 's': LEput('\\'); break;
    case 'n': LEput('\n'); break;
    case 't': LEput('\t'); break;
    case 'e': LEput(0x1b); break;
    case 'x':
      if (LEget(1))           return LEtoken(LEtUNTERMQ);
      if ((d1=hexval(LEc))<0) return LEtoken(LEtBADQ);
      if (LEget(1))           return LEtoken(LEtUNTERMQ);
      if ((d2=hexval(LEc))<0) return LEtoken(LEtBADQ);
      LEput( (d1<<4) | d2);
      break;
    default:
      return LEtoken(LEtBADQ);
    }
    goto quotcont;
  }
  LEput(LEc);
  goto quotcont;
Toper:
  LEtok= LEc;
  LEput(LEc);
  LEinpH++;
  return LEtoken(LEtok);
Tident:
  LEput(LEc);
  if (LEget(1)) return LEtoken(LEtIDENT);
  if (LEcls[LEc]==LEclsLETTER || LEcls[LEc]==LEclsDIGIT) goto Tident;
  return LEtoken(LEtIDENT);
Tnumber:
  LEput(LEc);
  LEsub= '9';
  if (LEc!='0') goto decnum; 
  if (LEget(1)) return LEtoken(LEtNUMBER);
  if (LEc=='b' || LEc=='B') goto binnum;
  if (LEc=='x' || LEc=='X') goto hexnum;
  if (LEc=='o') goto octnum;

  if (LEcls[LEc]!=LEclsDIGIT && LEc!='_') return LEtoken(LEtNUMBER);

decnum:
  if (LEget(1)) return LEtoken(LEtNUMBER);
  if (LEc>='0' && LEc<='9') { LEput(LEc); goto decnum; }
  if (LEc=='_') goto decnum; 
  return LEtoken(LEtNUMBER);

binnum:
  LEvallen= 0;
  LEsub= '1';
  if (LEget(1)) return LEtoken(LEtBADNUM); 
  if (LEc=='_') goto binnum;
  if (LEc!='0' && LEc!='1') return LEtoken(LEtBADNUM);
bincon:
  LEput(LEc);
binconnoput:
  if (LEget(1)) return LEtoken(LEtNUMBER);
  if (LEc=='0' || LEc=='1') goto bincon;
  if (LEc=='_') goto binconnoput;
  return LEtoken(LEtNUMBER);

hexnum:
  LEvallen= 0;
  LEsub= 'f';
  if (LEget(1)) return LEtoken(LEtBADNUM); 
  if (LEc=='_') goto hexnum;
  if (hexval(LEc)==-1) return LEtoken(LEtBADNUM);
hexcon:
  LEput(LEc);
hexconnoput:
  if (LEget(1)) return LEtoken(LEtNUMBER);
  if (hexval(LEc)!=-1) goto hexcon;
  if (LEc=='_') goto hexconnoput;
  return LEtoken(LEtNUMBER);

octnum:
  LEvallen= 0;
  LEsub= '7';
  if (LEget(1)) return LEtoken(LEtBADNUM); 
  if (LEc=='_') goto octnum;
  if (LEc<'0' || LEc>'7') return LEtoken(LEtBADNUM);
octcon:
  LEput(LEc);
octconnoput:
  if (LEget(1)) return LEtoken(LEtNUMBER);
  if (LEc>='0' && LEc<='7') goto octcon; 
  if (LEc=='_') goto octconnoput;
  return LEtoken(LEtNUMBER);
}

int LEpair(int lp, int rp)
{
  int q, count;
  LEvallen= 0;
  count= 1;
again:
  if (LEget(0)) return LEtoken(LEtEOF);
  if (LEc=='\n') { LElin++; LEput(LEc); LEinpH++; goto again; }

  if (LEc==lp) { count++; LEput(LEc); LEinpH++; goto again; }
  if (LEc==rp) { count--; LEinpH++; if (count==0) return LEtoken(0);
                 LEput('}'); goto again; }
  if (LEc=='#') 
  {
    LEput(LEc);
    while(!LEget(1))
    {
       LEput(LEc);
       if (LEc=='\n') { LEinpH++; goto again; }
    }
    return LEtoken(LEtEOF);
  }
  if (LEc=='\'' || LEc=='\"')
  {
    q= LEc;
    LEput(LEc);
    while(!LEget(1))
    {
      LEput(LEc);
      if (LEc==q) { LEinpH++; goto again; }
      if (LEc=='\\') { if (LEget(1)) return LEtoken(LEtEOF); LEput(LEc); }
    }
    return LEtoken(LEtEOF);
  }
  LEput(LEc);
  LEinpH++;
  goto again;
#undef LEc
}

int LEisIDENT(const char *v)
{
  if (LEtok!=LEtIDENT) return 0;
  if (strcmp(v, (char*) LEval)) return 0;
  return 1;
}

