#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "ptrbuf.h"
#include "proc.h"
#include "lexer.h"

void parse_error(char *msg)
{
  fprintf(stderr,"%s:%d %s\n", LEfil, LElin, msg);
  exit(1);
}

void parse_args(size_t *Rnargs, arg_t ***Rargs)
{
  int i,V;
  int type, COM,BAS,WID;
  char *k;
  ptrbuf_t *B;
  arg_t *A;
  B= ptrbuf_new(10);
  
  LEnext(); // skip the opening parenthesis

  while(LEtok!=')')
  {
    COM= '.';
    if (LEtok!=LEtIDENT) parse_error("identifier expected");
    k= (char*) LEval;
    if (k[0]=='a' || k[0]=='s') { COM= k[0]; k++; }
    if (!k[0]) parse_error("bad type");
    if (strchr("uif",k[0])) { BAS= k[0]; k++; }
                       else parse_error("bad type");
    if (!isdigit(k[0])) parse_error("bad type");
    for(i=0,V=0;isdigit(k[i]);i++) V= V*10 + k[i]-'0'; 
    if (k[i]!=0) parse_error("bad type");
    WID= V;
    if (BAS=='f') { if (COM=='s' || (WID!=32 && WID!=64))
                      parse_error("bad type"); }
    else { if (WID!=8 && WID!=16 && WID!=32)
                      parse_error("bad type"); }

    type= TYPE(COM,BAS,WID);

    LEnext();

    while(1)
    {
      if (LEtok!=LEtIDENT) parse_error("identifier expected");
      A= malloc(sizeof(*A));
      A->type= type;
      A->name= strdup((char*) LEval);
      ptrbuf_put(B, A);
      // printf("argument <%08x,%s>\n", type, LEval);
      LEnext();
      if (LEtok==')') break;
      if (LEtok==',') { LEnext(); continue; }
      if (LEtok==';') { LEnext(); break; }
      parse_error("syntax error");
    }
  }
  LEnext();
  ptrbuf_free(B, (void***) Rargs, Rnargs);
}

method_t* parse_method()
{
  method_t *M;
  if (!LEisIDENT("method")) 
    parse_error("internal error: 'method' is expected");
  LEnext();
  if (LEtok!=LEtIDENT)
    parse_error("identifier expected");
  M= malloc(sizeof(*M));
  M->name= strdup((char*) LEval);
  LEnext();
  if (LEtok!='(')
    parse_error("'(' expected");
  parse_args(&M->nargs, &M->args);
  return M;
}



void parse_file(char *fn, size_t *Rnprocs, proc_t ***Rprocs)
{
  proc_t *P;
  ptrbuf_t *methods;
  ptrbuf_t *procs;
  LEinit();
  LEfile(fn);
  LEnext();
  methods= NULL; P= NULL;
  procs= ptrbuf_new(10);
  
  while(LEtok!=LEtEOF)
  {
    if (LEisIDENT("process"))
    {
      LEnext();
      if (LEtok!=LEtIDENT) parse_error("identifier expected");
      if (methods)
        ptrbuf_free(methods, (void***) &P->methods, &P->nmethods);
      P= calloc(1,sizeof(*P));
      ptrbuf_put(procs, P);
      P->name= strdup((char*)LEval);
      methods= ptrbuf_new(20);
      LEnext();
      continue;
    }
    if (LEisIDENT("method"))
    {
      if (!P) parse_error("methods can't be defined outside processes");
      ptrbuf_put(methods, parse_method());    
      continue;
    }
    if (LEisIDENT("header"))
    {
      if (!P) parse_error("header files can't be defined outside processes");
      LEnext();
      if (LEtok!=LEtQUOTED) parse_error("quoted string expected");
      P->header_fn= strdup((char*)LEval);
      LEnext(); 
      continue;
    }
    if (LEisIDENT("source"))
    {
      if (!P) parse_error("source files can't be defined outside processes");
      LEnext();
      if (LEtok!=LEtQUOTED) parse_error("quoted string expected");
      P->source_fn= strdup((char*)LEval);
      LEnext(); 
      continue;
    }
    parse_error("unexpected token");
  }
  if (methods)
    ptrbuf_free(methods, (void***) &P->methods, &P->nmethods);
  ptrbuf_free(procs, (void***) Rprocs, Rnprocs);
}

