
#define TYPE(COM,BASE,WIDTH)  (((COM)<<16)|((BASE)<<8)|(WIDTH))
#define TYPEcom(X)  (((X)>>16)&0xff)
#define TYPEwid(X)  ((X)&0xff)
#define TYPEbas(X)  (((X)>>8)&0xff)

//
// types are used like this:  
//   TYPE('s','u',8) => string of bytes
//   TYPE('.','i',16) => single uint16_t
//   TYPE('a','f',32) => array of floats
// bases are:
//   u: unsigned int 
//   i: signed int
//   f: floating point
// composites are:
//   s: string (not for floats)
//   a: array
//   .: single value
// widths are:
//   8,16,32: integers
//     32,64: floating point
//
// When defining arguments, these are written as identifiers
// such as  f32 or su8 etc. 
//

typedef struct
{
  char *name;
  int type;
} arg_t;

typedef struct
{
  char *name;
  size_t nargs;
  arg_t **args;
  int code;
} method_t;
typedef struct
{
  char *name;
  char *source_fn;
  char *header_fn;
  FILE *source, *header;
  method_t **methods;
  size_t nmethods;
} proc_t;

