/*** IPF source file for process term ***/
#define _OBUFSIZE 12
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>


static uint8_t _OBUF[_OBUFSIZE];
static uint8_t *_IBUF;
static uint32_t _IBUFsize;

static int Osock=-1;
static int Oserv=-1;


#define PAD4(X)  (~4 & (4-((X)%4)))
#define PAD8(X)  (~8 & (8-((X)%8)))

static uint32_t strleni8(char *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}
static uint32_t strlenu8(unsigned char *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}
static uint32_t strleni16(int16_t *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}
static uint32_t strlenu16(uint16_t *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}
static uint32_t strleni32(int32_t *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}
static uint32_t strlenu32(uint32_t *p)
{int i;if(!p)return 0;for(i=0;p[i];i++);return i+1;}

static int close_sock()
{
  if (Osock==-1) return 1;
  shutdown(Osock, SHUT_RDWR);
  close(Osock);
  Osock= -1;
  return 1;
}

static int OSENDx(void *_buf, size_t len)
{
  char *buf;
  ssize_t R;
  if (Osock==-1) return 1;
  buf= _buf;
  while(len)
  {
     R= send(Osock, buf, len, MSG_NOSIGNAL);
     if (R<0)
     {
       if (errno==EINTR) continue;
       return close_sock();
     }
     buf += R;
     len -= R;
  }
  return 0;
}

static int OSEND(void *buf, size_t len, size_t off, int pad)
{
  static uint8_t padb[8];
  int pa;
  pa= pad - (off%pad);
  if (pa!=pad)
  {
    if (OSENDx(padb, pa)) return -1;
    off+= pa;
  }
  if (OSENDx(buf,len)) return -1;
  return off + len;
}

static int IRECV(void *_buf, size_t len)
{
  char *buf;
  ssize_t R;
  if (Osock==-1) return 1;
  buf= _buf;
  while(len)
  {
     R= recv(Osock, buf, len, 0);
     if (R<0)
     {
       if (errno==EINTR) continue;
       return close_sock();
     }
     if (R==0) return close_sock();
     buf += R;
     len -= R;
  }
  return 0;
}

static void _DISPATCH();

int IPFread()
{
  uint32_t L;
  if (Osock==-1) return 1;
  if (IRECV(&L,4)) return 1;
  _IBUF= malloc(L);
  if (IRECV(_IBUF+4,L-4)) { free(_IBUF); return 1; }
  _IBUFsize= L;
  _DISPATCH();
  free(_IBUF);
  return 0;
}


static int Iserver(int port)
{
  int yes= 1;
  int ec;
  struct sockaddr_in addr;
  Oserv= socket(PF_INET, SOCK_STREAM, 0);
  if (Oserv<0) return 1;
  setsockopt(Oserv, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  addr.sin_port= htons( (short) port);
  addr.sin_addr.s_addr= INADDR_ANY;
  if (!bind(Oserv, (struct sockaddr*) &addr,sizeof(addr)) &&
      !listen(Oserv, 5)) 
     return 0;
  ec= errno;
  close(Oserv);
  errno= ec;
  Oserv= -1;
  return 0;
}

static int Iaccept()
{
  int sock;
  if (Oserv==-1) return 1;
  sock= accept(Oserv, NULL, NULL);
  if (sock<0) return 1;
  if (Osock!=-1) close_sock();
  Osock= sock;
  return 0;
}

int IPFconnect(char *host,int port) 
{
  struct sockaddr_in addr;
  int len= sizeof(addr);
  int ec;

  Osock= socket(PF_INET, SOCK_STREAM, 0);
  if (Osock<0) return 1;
  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  addr.sin_port= htons( (short) port);
  inet_aton(host,&addr.sin_addr);
  if (connect(Osock,(struct sockaddr*) &addr, len))
  {
    ec= errno;
    close_sock();
    errno= ec;
    return -1;
  }
  return 0;
}

int IPFserver(int port)
{
  if (Iserver(port)) return -1;
  return Oserv;
}

int IPFaccept()
{
  if (Iaccept()) return -1;
  return Osock;
}

int IPFserverSOCK() { return Oserv; }
int IPFdataSOCK() { return Osock; }


void *__IPF_unused[]= 
{
  strleni8, strlenu8,
  strleni16, strlenu16,
  strleni32, strlenu32,
};
/************************************************************************
 *   Functions implemented on this side. Implement these functions      *
 *   in this program. The peer will cause these to be called.           *
 ***********************************************************************/
void result(int32_t b) ;
void message(char* a) ;
void getinput() ;
static void _DISP2()
{
static uint8_t *_INP;
  int32_t  b;
  _INP= _IBUF+8; // skip length+mcode
  b= *(int32_t*)_INP; _INP+= 4;
  result(b);
}
static void _DISP3()
{
static uint8_t *_INP;
  char*  a;
  uint32_t La;
  _INP= _IBUF+8; // skip length+mcode
  La= *(uint32_t*)_INP; _INP+= 4;
  _INP+= PAD4(_INP-_IBUF); a= (char*) _INP; _INP+= La*1;
  message(a);
}
static void _DISP4()
{
  getinput();
}
static void _DISPATCH()
{
  uint32_t mcode;
  mcode= *(uint32_t*) (_IBUF+4);
  switch(mcode)
  {
    case 2: _DISP2(); break;
    case 3: _DISP3(); break;
    case 4: _DISP4(); break;
  }
}
int input(int32_t a) 
{
  uint32_t _LEN;
  uint8_t *_OUT;
  uint32_t _OFF;
  _LEN= 8; // length and mcode fields 
  _LEN+= 4; // a
  _OUT= _OBUF;
  *(uint32_t*) _OUT= _LEN; _OUT+= 4;
  *(uint32_t*) _OUT= 1; _OUT+= 4;
  *(int32_t*) _OUT= a; _OUT+= 4;
  _OFF= _OUT-_OBUF;
  if (OSEND(_OBUF, _OFF, 0, 4)<0) return 1;
  return 0;
}
